# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.  This
was written to work with FSL version 5.0.4.
"""
import os
import numpy as np
import nibabel as nb
import warnings

from ...utils.filemanip import split_filename, fname_presuffix

from ..base import traits, TraitedSpec, InputMultiPath, File, isdefined
from .base import FSLCommand, FSLCommandInputSpec, Info


class PrepareFieldmapInputSpec(FSLCommandInputSpec):
    scanner = traits.String(
        "SIEMENS", argstr="%s", position=1, desc="must be SIEMENS", usedefault=True
    )
    in_phase = File(
        exists=True,
        argstr="%s",
        position=2,
        mandatory=True,
        desc=(
            "Phase difference map, in SIEMENS format range from " "0-4096 or 0-8192)"
        ),
    )
    in_magnitude = File(
        exists=True,
        argstr="%s",
        position=3,
        mandatory=True,
        desc="Magnitude difference map, brain extracted",
    )
    delta_TE = traits.Float(
        2.46,
        usedefault=True,
        mandatory=True,
        argstr="%f",
        position=-2,
        desc=(
            "echo time difference of the "
            "fieldmap sequence in ms. (usually 2.46ms in"
            " Siemens)"
        ),
    )
    nocheck = traits.Bool(
        False,
        position=-1,
        argstr="--nocheck",
        usedefault=True,
        desc=("do not perform sanity checks for image " "size/range/dimensions"),
    )
    out_fieldmap = File(
        argstr="%s", position=4, desc="output name for prepared fieldmap"
    )


class PrepareFieldmapOutputSpec(TraitedSpec):
    out_fieldmap = File(exists=True, desc="output name for prepared fieldmap")


class PrepareFieldmap(FSLCommand):
    """
    Interface for the fsl_prepare_fieldmap script (FSL 5.0)

    Prepares a fieldmap suitable for FEAT from SIEMENS data - saves output in
    rad/s format (e.g. ```fsl_prepare_fieldmap SIEMENS
    images_3_gre_field_mapping images_4_gre_field_mapping fmap_rads 2.65```).


    Examples
    --------

    >>> from nipype.interfaces.fsl import PrepareFieldmap
    >>> prepare = PrepareFieldmap()
    >>> prepare.inputs.in_phase = "phase.nii"
    >>> prepare.inputs.in_magnitude = "magnitude.nii"
    >>> prepare.inputs.output_type = "NIFTI_GZ"
    >>> prepare.cmdline # doctest: +ELLIPSIS
    'fsl_prepare_fieldmap SIEMENS phase.nii magnitude.nii \
.../phase_fslprepared.nii.gz 2.460000'
    >>> res = prepare.run() # doctest: +SKIP


    """

    _cmd = "fsl_prepare_fieldmap"
    input_spec = PrepareFieldmapInputSpec
    output_spec = PrepareFieldmapOutputSpec

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []

        if not isdefined(self.inputs.out_fieldmap):
            self.inputs.out_fieldmap = self._gen_fname(
                self.inputs.in_phase, suffix="_fslprepared"
            )

        if not isdefined(self.inputs.nocheck) or not self.inputs.nocheck:
            skip += ["nocheck"]

        return super(PrepareFieldmap, self)._parse_inputs(skip=skip)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_fieldmap"] = self.inputs.out_fieldmap
        return outputs

    def _run_interface(self, runtime):
        runtime = super(PrepareFieldmap, self)._run_interface(runtime)

        if runtime.returncode == 0:
            out_file = self.inputs.out_fieldmap
            im = nb.load(out_file)
            dumb_img = nb.Nifti1Image(np.zeros(im.shape), im.affine, im.header)
            out_nii = nb.funcs.concat_images((im, dumb_img))
            nb.save(out_nii, out_file)

        return runtime


class TOPUPInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        desc="name of 4D file with images",
        argstr="--imain=%s",
    )
    encoding_file = File(
        exists=True,
        mandatory=True,
        xor=["encoding_direction"],
        desc="name of text file with PE directions/times",
        argstr="--datain=%s",
    )
    encoding_direction = traits.List(
        traits.Enum("y", "x", "z", "x-", "y-", "z-"),
        mandatory=True,
        xor=["encoding_file"],
        requires=["readout_times"],
        argstr="--datain=%s",
        desc=("encoding direction for automatic " "generation of encoding_file"),
    )
    readout_times = InputMultiPath(
        traits.Float,
        requires=["encoding_direction"],
        xor=["encoding_file"],
        mandatory=True,
        desc=("readout times (dwell times by # " "phase-encode steps minus 1)"),
    )
    out_base = File(
        desc=(
            "base-name of output files (spline "
            "coefficients (Hz) and movement parameters)"
        ),
        name_source=["in_file"],
        name_template="%s_base",
        argstr="--out=%s",
        hash_files=False,
    )
    out_field = File(
        argstr="--fout=%s",
        hash_files=False,
        name_source=["in_file"],
        name_template="%s_field",
        desc="name of image file with field (Hz)",
    )
    out_warp_prefix = traits.Str(
        "warpfield",
        argstr="--dfout=%s",
        hash_files=False,
        desc="prefix for the warpfield images (in mm)",
        usedefault=True,
    )
    out_mat_prefix = traits.Str(
        "xfm",
        argstr="--rbmout=%s",
        hash_files=False,
        desc="prefix for the realignment matrices",
        usedefault=True,
    )
    out_jac_prefix = traits.Str(
        "jac",
        argstr="--jacout=%s",
        hash_files=False,
        desc="prefix for the warpfield images",
        usedefault=True,
    )
    out_corrected = File(
        argstr="--iout=%s",
        hash_files=False,
        name_source=["in_file"],
        name_template="%s_corrected",
        desc="name of 4D image file with unwarped images",
    )
    out_logfile = File(
        argstr="--logout=%s",
        desc="name of log-file",
        name_source=["in_file"],
        name_template="%s_topup.log",
        keep_extension=True,
        hash_files=False,
    )

    # TODO: the following traits admit values separated by commas, one value
    # per registration level inside topup.
    warp_res = traits.Float(
        argstr="--warpres=%f",
        desc=(
            "(approximate) resolution (in mm) of warp "
            "basis for the different sub-sampling levels"
        ),
    )
    subsamp = traits.Int(argstr="--subsamp=%d", desc="sub-sampling scheme")
    fwhm = traits.Float(
        argstr="--fwhm=%f", desc="FWHM (in mm) of gaussian smoothing kernel"
    )
    config = traits.String(
        "b02b0.cnf",
        argstr="--config=%s",
        usedefault=True,
        desc=("Name of config file specifying command line " "arguments"),
    )
    max_iter = traits.Int(argstr="--miter=%d", desc="max # of non-linear iterations")
    reg_lambda = traits.Float(
        argstr="--lambda=%0.f",
        desc=(
            "Weight of regularisation, default "
            "depending on --ssqlambda and --regmod switches."
        ),
    )
    ssqlambda = traits.Enum(
        1,
        0,
        argstr="--ssqlambda=%d",
        desc=(
            "Weight lambda by the current value of the "
            "ssd. If used (=1), the effective weight of "
            "regularisation term becomes higher for the "
            "initial iterations, therefore initial steps"
            " are a little smoother than they would "
            "without weighting. This reduces the "
            "risk of finding a local minimum."
        ),
    )
    regmod = traits.Enum(
        "bending_energy",
        "membrane_energy",
        argstr="--regmod=%s",
        desc=(
            "Regularisation term implementation. Defaults "
            "to bending_energy. Note that the two functions"
            " have vastly different scales. The membrane "
            "energy is based on the first derivatives and "
            "the bending energy on the second derivatives. "
            "The second derivatives will typically be much "
            "smaller than the first derivatives, so input "
            "lambda will have to be larger for "
            "bending_energy to yield approximately the same"
            " level of regularisation."
        ),
    )
    estmov = traits.Enum(1, 0, argstr="--estmov=%d", desc="estimate movements if set")
    minmet = traits.Enum(
        0,
        1,
        argstr="--minmet=%d",
        desc=(
            "Minimisation method 0=Levenberg-Marquardt, " "1=Scaled Conjugate Gradient"
        ),
    )
    splineorder = traits.Int(
        argstr="--splineorder=%d",
        desc=("order of spline, 2->Qadratic spline, " "3->Cubic spline"),
    )
    numprec = traits.Enum(
        "double",
        "float",
        argstr="--numprec=%s",
        desc=("Precision for representing Hessian, double " "or float."),
    )
    interp = traits.Enum(
        "spline",
        "linear",
        argstr="--interp=%s",
        desc="Image interpolation model, linear or spline.",
    )
    scale = traits.Enum(
        0,
        1,
        argstr="--scale=%d",
        desc=("If set (=1), the images are individually scaled" " to a common mean"),
    )
    regrid = traits.Enum(
        1,
        0,
        argstr="--regrid=%d",
        desc=("If set (=1), the calculations are done in a " "different grid"),
    )


class TOPUPOutputSpec(TraitedSpec):
    out_fieldcoef = File(exists=True, desc="file containing the field coefficients")
    out_movpar = File(exists=True, desc="movpar.txt output file")
    out_enc_file = File(desc="encoding directions file output for applytopup")
    out_field = File(desc="name of image file with field (Hz)")
    out_warps = traits.List(File(exists=True), desc="warpfield images")
    out_jacs = traits.List(File(exists=True), desc="Jacobian images")
    out_mats = traits.List(File(exists=True), desc="realignment matrices")
    out_corrected = File(desc="name of 4D image file with unwarped images")
    out_logfile = File(desc="name of log-file")


class TOPUP(FSLCommand):
    """
    Interface for FSL topup, a tool for estimating and correcting
    susceptibility induced distortions. See FSL documentation for
    `reference <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/TOPUP>`_,
    `usage examples
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/topup/ExampleTopupFollowedByApplytopup>`_,
    and `exemplary config files
    <https://github.com/ahheckel/FSL-scripts/blob/master/rsc/fsl/fsl4/topup/b02b0.cnf>`_.

    Examples
    --------

    >>> from nipype.interfaces.fsl import TOPUP
    >>> topup = TOPUP()
    >>> topup.inputs.in_file = "b0_b0rev.nii"
    >>> topup.inputs.encoding_file = "topup_encoding.txt"
    >>> topup.inputs.output_type = "NIFTI_GZ"
    >>> topup.cmdline # doctest: +ELLIPSIS
    'topup --config=b02b0.cnf --datain=topup_encoding.txt \
--imain=b0_b0rev.nii --out=b0_b0rev_base --iout=b0_b0rev_corrected.nii.gz \
--fout=b0_b0rev_field.nii.gz --jacout=jac --logout=b0_b0rev_topup.log \
--rbmout=xfm --dfout=warpfield'
    >>> res = topup.run() # doctest: +SKIP

    """

    _cmd = "topup"
    input_spec = TOPUPInputSpec
    output_spec = TOPUPOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "encoding_direction":
            return trait_spec.argstr % self._generate_encfile()
        if name == "out_base":
            path, name, ext = split_filename(value)
            if path != "":
                if not os.path.exists(path):
                    raise ValueError("out_base path must exist if provided")
        return super(TOPUP, self)._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = super(TOPUP, self)._list_outputs()
        del outputs["out_base"]
        base_path = None
        if isdefined(self.inputs.out_base):
            base_path, base, _ = split_filename(self.inputs.out_base)
            if base_path == "":
                base_path = None
        else:
            base = split_filename(self.inputs.in_file)[1] + "_base"
        outputs["out_fieldcoef"] = self._gen_fname(
            base, suffix="_fieldcoef", cwd=base_path
        )
        outputs["out_movpar"] = self._gen_fname(
            base, suffix="_movpar", ext=".txt", cwd=base_path
        )

        n_vols = nb.load(self.inputs.in_file).shape[-1]
        ext = Info.output_type_to_ext(self.inputs.output_type)
        fmt = os.path.abspath("{prefix}_{i:02d}{ext}").format
        outputs["out_warps"] = [
            fmt(prefix=self.inputs.out_warp_prefix, i=i, ext=ext)
            for i in range(1, n_vols + 1)
        ]
        outputs["out_jacs"] = [
            fmt(prefix=self.inputs.out_jac_prefix, i=i, ext=ext)
            for i in range(1, n_vols + 1)
        ]
        outputs["out_mats"] = [
            fmt(prefix=self.inputs.out_mat_prefix, i=i, ext=".mat")
            for i in range(1, n_vols + 1)
        ]

        if isdefined(self.inputs.encoding_direction):
            outputs["out_enc_file"] = self._get_encfilename()
        return outputs

    def _get_encfilename(self):
        out_file = os.path.join(
            os.getcwd(), ("%s_encfile.txt" % split_filename(self.inputs.in_file)[1])
        )
        return out_file

    def _generate_encfile(self):
        """Generate a topup compatible encoding file based on given directions"""
        out_file = self._get_encfilename()
        durations = self.inputs.readout_times
        if len(self.inputs.encoding_direction) != len(durations):
            if len(self.inputs.readout_times) != 1:
                raise ValueError(
                    (
                        "Readout time must be a float or match the"
                        "length of encoding directions"
                    )
                )
            durations = durations * len(self.inputs.encoding_direction)

        lines = []
        for idx, encdir in enumerate(self.inputs.encoding_direction):
            direction = 1.0
            if encdir.endswith("-"):
                direction = -1.0
            line = [
                float(val[0] == encdir[0]) * direction for val in ["x", "y", "z"]
            ] + [durations[idx]]
            lines.append(line)
        np.savetxt(out_file, np.array(lines), fmt=b"%d %d %d %.8f")
        return out_file

    def _overload_extension(self, value, name=None):
        if name == "out_base":
            return value
        return super(TOPUP, self)._overload_extension(value, name)


class ApplyTOPUPInputSpec(FSLCommandInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc="name of file with images",
        argstr="--imain=%s",
        sep=",",
    )
    encoding_file = File(
        exists=True,
        mandatory=True,
        desc="name of text file with PE directions/times",
        argstr="--datain=%s",
    )
    in_index = traits.List(
        traits.Int,
        argstr="--inindex=%s",
        sep=",",
        desc="comma separated list of indices corresponding to --datain",
    )
    in_topup_fieldcoef = File(
        exists=True,
        argstr="--topup=%s",
        copyfile=False,
        requires=["in_topup_movpar"],
        desc=("topup file containing the field " "coefficients"),
    )
    in_topup_movpar = File(
        exists=True,
        requires=["in_topup_fieldcoef"],
        copyfile=False,
        desc="topup movpar.txt file",
    )
    out_corrected = File(
        desc="output (warped) image",
        name_source=["in_files"],
        name_template="%s_corrected",
        argstr="--out=%s",
    )
    method = traits.Enum(
        "jac",
        "lsr",
        argstr="--method=%s",
        desc=("use jacobian modulation (jac) or least-squares" " resampling (lsr)"),
    )
    interp = traits.Enum(
        "trilinear", "spline", argstr="--interp=%s", desc="interpolation method"
    )
    datatype = traits.Enum(
        "char",
        "short",
        "int",
        "float",
        "double",
        argstr="-d=%s",
        desc="force output data type",
    )


class ApplyTOPUPOutputSpec(TraitedSpec):
    out_corrected = File(
        exists=True, desc=("name of 4D image file with " "unwarped images")
    )


class ApplyTOPUP(FSLCommand):
    """
    Interface for FSL topup, a tool for estimating and correcting
    susceptibility induced distortions.
    `General reference
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/topup/ApplytopupUsersGuide>`_
    and `use example
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/topup/ExampleTopupFollowedByApplytopup>`_.


    Examples
    --------

    >>> from nipype.interfaces.fsl import ApplyTOPUP
    >>> applytopup = ApplyTOPUP()
    >>> applytopup.inputs.in_files = ["epi.nii", "epi_rev.nii"]
    >>> applytopup.inputs.encoding_file = "topup_encoding.txt"
    >>> applytopup.inputs.in_topup_fieldcoef = "topup_fieldcoef.nii.gz"
    >>> applytopup.inputs.in_topup_movpar = "topup_movpar.txt"
    >>> applytopup.inputs.output_type = "NIFTI_GZ"
    >>> applytopup.cmdline # doctest: +ELLIPSIS
    'applytopup --datain=topup_encoding.txt --imain=epi.nii,epi_rev.nii \
--inindex=1,2 --topup=topup --out=epi_corrected.nii.gz'
    >>> res = applytopup.run() # doctest: +SKIP

    """

    _cmd = "applytopup"
    input_spec = ApplyTOPUPInputSpec
    output_spec = ApplyTOPUPOutputSpec

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []

        # If not defined, assume index are the first N entries in the
        # parameters file, for N input images.
        if not isdefined(self.inputs.in_index):
            self.inputs.in_index = list(range(1, len(self.inputs.in_files) + 1))

        return super(ApplyTOPUP, self)._parse_inputs(skip=skip)

    def _format_arg(self, name, spec, value):
        if name == "in_topup_fieldcoef":
            return spec.argstr % value.split("_fieldcoef")[0]
        return super(ApplyTOPUP, self)._format_arg(name, spec, value)


class EddyInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="--imain=%s",
        desc="File containing all the images to estimate distortions for",
    )
    in_mask = File(
        exists=True, mandatory=True, argstr="--mask=%s", desc="Mask to indicate brain"
    )
    in_index = File(
        exists=True,
        mandatory=True,
        argstr="--index=%s",
        desc="File containing indices for all volumes in --imain "
        "into --acqp and --topup",
    )
    in_acqp = File(
        exists=True,
        mandatory=True,
        argstr="--acqp=%s",
        desc="File containing acquisition parameters",
    )
    in_bvec = File(
        exists=True,
        mandatory=True,
        argstr="--bvecs=%s",
        desc="File containing the b-vectors for all volumes in --imain",
    )
    in_bval = File(
        exists=True,
        mandatory=True,
        argstr="--bvals=%s",
        desc="File containing the b-values for all volumes in --imain",
    )
    out_base = traits.Str(
        default_value="eddy_corrected",
        usedefault=True,
        argstr="--out=%s",
        desc="Basename for output image",
    )
    session = File(
        exists=True,
        argstr="--session=%s",
        desc="File containing session indices for all volumes in --imain",
    )
    in_topup_fieldcoef = File(
        exists=True,
        argstr="--topup=%s",
        requires=["in_topup_movpar"],
        desc="Topup results file containing the field coefficients",
    )
    in_topup_movpar = File(
        exists=True,
        requires=["in_topup_fieldcoef"],
        desc="Topup results file containing the movement parameters (movpar.txt)",
    )
    field = File(
        exists=True, argstr="--field=%s", desc="Non-topup derived fieldmap scaled in Hz"
    )
    field_mat = File(
        exists=True,
        argstr="--field_mat=%s",
        desc="Matrix specifying the relative positions of the fieldmap, "
        "--field, and the first volume of the input file, --imain",
    )

    flm = traits.Enum(
        "quadratic",
        "linear",
        "cubic",
        usedefault=True,
        argstr="--flm=%s",
        desc="First level EC model",
    )
    slm = traits.Enum(
        "none",
        "linear",
        "quadratic",
        usedefault=True,
        argstr="--slm=%s",
        desc="Second level EC model",
    )
    fep = traits.Bool(
        False, argstr="--fep", desc="Fill empty planes in x- or y-directions"
    )
    initrand = traits.Bool(
        False,
        argstr="--initrand",
        desc="Resets rand for when selecting voxels",
        min_ver="5.0.10",
    )
    interp = traits.Enum(
        "spline",
        "trilinear",
        usedefault=True,
        argstr="--interp=%s",
        desc="Interpolation model for estimation step",
    )
    nvoxhp = traits.Int(
        default_value=1000,
        usedefault=True,
        argstr="--nvoxhp=%s",
        desc="# of voxels used to estimate the hyperparameters",
    )
    fudge_factor = traits.Float(
        default_value=10.0,
        usedefault=True,
        argstr="--ff=%s",
        desc="Fudge factor for hyperparameter error variance",
    )
    dont_sep_offs_move = traits.Bool(
        False,
        argstr="--dont_sep_offs_move",
        desc="Do NOT attempt to separate field offset from subject movement",
    )
    dont_peas = traits.Bool(
        False,
        argstr="--dont_peas",
        desc="Do NOT perform a post-eddy alignment of shells",
    )
    fwhm = traits.Float(
        desc="FWHM for conditioning filter when estimating the parameters",
        argstr="--fwhm=%s",
    )
    niter = traits.Int(
        5, usedefault=True, argstr="--niter=%s", desc="Number of iterations"
    )
    method = traits.Enum(
        "jac",
        "lsr",
        usedefault=True,
        argstr="--resamp=%s",
        desc="Final resampling method (jacobian/least squares)",
    )

    repol = traits.Bool(
        False, argstr="--repol", desc="Detect and replace outlier slices"
    )
    outlier_nstd = traits.Int(
        argstr="--ol_nstd",
        desc="Number of std off to qualify as outlier",
        requires=["repol"],
        min_ver="5.0.10",
    )
    outlier_nvox = traits.Int(
        argstr="--ol_nvox",
        desc="Min # of voxels in a slice for inclusion in outlier detection",
        requires=["repol"],
        min_ver="5.0.10",
    )
    outlier_type = traits.Enum(
        "sw",
        "gw",
        "both",
        argstr="--ol_type",
        desc="Type of outliers, slicewise (sw), groupwise (gw) or both (both)",
        requires=["repol"],
        min_ver="5.0.10",
    )
    outlier_pos = traits.Bool(
        False,
        argstr="--ol_pos",
        desc="Consider both positive and negative outliers if set",
        requires=["repol"],
        min_ver="5.0.10",
    )
    outlier_sqr = traits.Bool(
        False,
        argstr="--ol_sqr",
        desc="Consider outliers among sums-of-squared differences if set",
        requires=["repol"],
        min_ver="5.0.10",
    )
    multiband_factor = traits.Int(
        argstr="--mb=%s", desc="Multi-band factor", min_ver="5.0.10"
    )
    multiband_offset = traits.Enum(
        0,
        1,
        -1,
        argstr="--mb_offs=%d",
        desc="Multi-band offset (-1 if bottom slice removed, 1 if top slice removed",
        requires=["multiband_factor"],
        min_ver="5.0.10",
    )

    mporder = traits.Int(
        argstr="--mporder=%s",
        desc="Order of slice-to-vol movement model",
        requires=["use_cuda"],
        min_ver="5.0.11",
    )
    slice2vol_niter = traits.Int(
        argstr="--s2v_niter=%d",
        desc="Number of iterations for slice-to-vol",
        requires=["mporder"],
        min_ver="5.0.11",
    )
    slice2vol_lambda = traits.Int(
        argstr="--s2v_lambda=%d",
        desc="Regularisation weight for slice-to-vol movement (reasonable range 1-10)",
        requires=["mporder"],
        min_ver="5.0.11",
    )
    slice2vol_interp = traits.Enum(
        "trilinear",
        "spline",
        argstr="--s2v_interp=%s",
        desc="Slice-to-vol interpolation model for estimation step",
        requires=["mporder"],
        min_ver="5.0.11",
    )
    slice_order = traits.File(
        exists=True,
        argstr="--slspec=%s",
        desc="Name of text file completely specifying slice/group acquisition",
        requires=["mporder"],
        xor=["json"],
        min_ver="5.0.11",
    )
    json = traits.File(
        exists=True,
        argstr="--json=%s",
        desc="Name of .json text file with information about slice timing",
        requires=["mporder"],
        xor=["slice_order"],
        min_ver="6.0.1",
    )

    estimate_move_by_susceptibility = traits.Bool(
        False,
        argstr="--estimate_move_by_susceptibility",
        desc="Estimate how susceptibility field changes with subject movement",
        min_ver="6.0.1",
    )
    mbs_niter = traits.Int(
        argstr="--mbs_niter=%s",
        desc="Number of iterations for MBS estimation",
        requires=["estimate_move_by_susceptibility"],
        min_ver="6.0.1",
    )
    mbs_lambda = traits.Int(
        argstr="--mbs_lambda=%s",
        desc="Weighting of regularisation for MBS estimation",
        requires=["estimate_move_by_susceptibility"],
        min_ver="6.0.1",
    )
    mbs_ksp = traits.Int(
        argstr="--mbs_ksp=%smm",
        desc="Knot-spacing for MBS field estimation",
        requires=["estimate_move_by_susceptibility"],
        min_ver="6.0.1",
    )

    num_threads = traits.Int(
        1, usedefault=True, nohash=True, desc="Number of openmp threads to use"
    )
    is_shelled = traits.Bool(
        False,
        argstr="--data_is_shelled",
        desc="Override internal check to ensure that date are acquired "
        "on a set of b-value shells",
    )

    use_cuda = traits.Bool(False, desc="Run eddy using cuda gpu")
    cnr_maps = traits.Bool(
        False, desc="Output CNR-Maps", argstr="--cnr_maps", min_ver="5.0.10"
    )
    residuals = traits.Bool(
        False, desc="Output Residuals", argstr="--residuals", min_ver="5.0.10"
    )


class EddyOutputSpec(TraitedSpec):
    out_corrected = File(
        exists=True, desc="4D image file containing all the corrected volumes"
    )
    out_parameter = File(
        exists=True,
        desc="Text file with parameters defining the field and movement for each scan",
    )
    out_rotated_bvecs = File(
        exists=True, desc="File containing rotated b-values for all volumes"
    )
    out_movement_rms = File(
        exists=True, desc="Summary of the 'total movement' in each volume"
    )
    out_restricted_movement_rms = File(
        exists=True,
        desc="Summary of the 'total movement' in each volume "
        "disregarding translation in the PE direction",
    )
    out_shell_alignment_parameters = File(
        exists=True,
        desc="Text file containing rigid body movement parameters "
        "between the different shells as estimated by a "
        "post-hoc mutual information based registration",
    )
    out_shell_pe_translation_parameters = File(
        exists=True,
        desc="Text file containing translation along the PE-direction "
        "between the different shells as estimated by a "
        "post-hoc mutual information based registration",
    )
    out_shell_pe_translation_parameters = File(
        exists=True,
        desc="Text file containing translation along the PE-direction "
        "between the different shells as estimated by a "
        "post-hoc mutual information based registration",
    )
    out_outlier_map = File(
        exists=True,
        desc="Matrix where rows represent volumes and columns represent "
        'slices. "0" indicates that scan-slice is not an outlier '
        'and "1" indicates that it is',
    )
    out_outlier_n_stdev_map = File(
        exists=True,
        desc="Matrix where rows represent volumes and columns represent "
        "slices. Values indicate number of standard deviations off the "
        "mean difference between observation and prediction is",
    )
    out_outlier_n_sqr_stdev_map = File(
        exists=True,
        desc="Matrix where rows represent volumes and columns represent "
        "slices. Values indicate number of standard deivations off the "
        "square root of the mean squared difference between observation "
        "and prediction is",
    )
    out_outlier_report = File(
        exists=True,
        desc="Text file with a plain language report on what "
        "outlier slices eddy has found",
    )
    out_outlier_free = File(
        exists=True,
        desc="4D image file not corrected for susceptibility or eddy-"
        "current distortions or subject movement but with outlier "
        "slices replaced",
    )
    out_movement_over_time = File(
        exists=True,
        desc="Text file containing translations (mm) and rotations "
        "(radians) for each excitation",
    )
    out_cnr_maps = File(exists=True, desc="path/name of file with the cnr_maps")
    out_residuals = File(exists=True, desc="path/name of file with the residuals")


class Eddy(FSLCommand):
    """
    Interface for FSL eddy, a tool for estimating and correcting eddy
    currents induced distortions. `User guide
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/UsersGuide>`__ and
    `more info regarding acqp file
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/Faq#How_do_I_know_what_to_put_into_my_--acqp_file.3F>`_.

    Examples
    --------

    >>> from nipype.interfaces.fsl import Eddy

    Running eddy on a CPU using OpenMP:
    >>> eddy = Eddy()
    >>> eddy.inputs.in_file = 'epi.nii'
    >>> eddy.inputs.in_mask  = 'epi_mask.nii'
    >>> eddy.inputs.in_index = 'epi_index.txt'
    >>> eddy.inputs.in_acqp  = 'epi_acqp.txt'
    >>> eddy.inputs.in_bvec  = 'bvecs.scheme'
    >>> eddy.inputs.in_bval  = 'bvals.scheme'
    >>> eddy.cmdline          # doctest: +ELLIPSIS
    'eddy_openmp --flm=quadratic --ff=10.0 \
--acqp=epi_acqp.txt --bvals=bvals.scheme --bvecs=bvecs.scheme \
--imain=epi.nii --index=epi_index.txt --mask=epi_mask.nii \
--interp=spline --resamp=jac --niter=5 --nvoxhp=1000 \
--out=.../eddy_corrected --slm=none'

    Running eddy on an Nvidia GPU using cuda:
    >>> eddy.inputs.use_cuda = True
    >>> eddy.cmdline # doctest: +ELLIPSIS
    'eddy_cuda --flm=quadratic --ff=10.0 \
--acqp=epi_acqp.txt --bvals=bvals.scheme --bvecs=bvecs.scheme \
--imain=epi.nii --index=epi_index.txt --mask=epi_mask.nii \
--interp=spline --resamp=jac --niter=5 --nvoxhp=1000 \
--out=.../eddy_corrected --slm=none'

    Running eddy with slice-to-volume motion correction:
    >>> eddy.inputs.mporder = 6
    >>> eddy.inputs.slice2vol_niter = 5
    >>> eddy.inputs.slice2vol_lambda = 1
    >>> eddy.inputs.slice2vol_interp = 'trilinear'
    >>> eddy.inputs.slice_order = 'epi_slspec.txt'
    >>> eddy.cmdline          # doctest: +ELLIPSIS
    'eddy_cuda --flm=quadratic --ff=10.0 \
--acqp=epi_acqp.txt --bvals=bvals.scheme --bvecs=bvecs.scheme \
--imain=epi.nii --index=epi_index.txt --mask=epi_mask.nii \
--interp=spline --resamp=jac --mporder=6 --niter=5 --nvoxhp=1000 \
--out=.../eddy_corrected --s2v_interp=trilinear --s2v_lambda=1 \
--s2v_niter=5 --slspec=epi_slspec.txt --slm=none'
    >>> res = eddy.run()     # doctest: +SKIP

    """

    _cmd = "eddy_openmp"
    input_spec = EddyInputSpec
    output_spec = EddyOutputSpec

    _num_threads = 1

    def __init__(self, **inputs):
        super(Eddy, self).__init__(**inputs)
        self.inputs.on_trait_change(self._num_threads_update, "num_threads")
        if not isdefined(self.inputs.num_threads):
            self.inputs.num_threads = self._num_threads
        else:
            self._num_threads_update()
        self.inputs.on_trait_change(self._use_cuda, "use_cuda")
        if isdefined(self.inputs.use_cuda):
            self._use_cuda()

    def _num_threads_update(self):
        self._num_threads = self.inputs.num_threads
        if not isdefined(self.inputs.num_threads):
            if "OMP_NUM_THREADS" in self.inputs.environ:
                del self.inputs.environ["OMP_NUM_THREADS"]
        else:
            self.inputs.environ["OMP_NUM_THREADS"] = str(self.inputs.num_threads)

    def _use_cuda(self):
        self._cmd = "eddy_cuda" if self.inputs.use_cuda else "eddy_openmp"

    def _run_interface(self, runtime):
        # If 'eddy_openmp' is missing, use 'eddy'
        FSLDIR = os.getenv("FSLDIR", "")
        cmd = self._cmd
        if all(
            (
                FSLDIR != "",
                cmd == "eddy_openmp",
                not os.path.exists(os.path.join(FSLDIR, "bin", cmd)),
            )
        ):
            self._cmd = "eddy"
        runtime = super(Eddy, self)._run_interface(runtime)

        # Restore command to avoid side-effects
        self._cmd = cmd
        return runtime

    def _format_arg(self, name, spec, value):
        if name == "in_topup_fieldcoef":
            return spec.argstr % value.split("_fieldcoef")[0]
        if name == "field":
            return spec.argstr % fname_presuffix(value, use_ext=False)
        if name == "out_base":
            return spec.argstr % os.path.abspath(value)
        return super(Eddy, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_corrected"] = os.path.abspath("%s.nii.gz" % self.inputs.out_base)
        outputs["out_parameter"] = os.path.abspath(
            "%s.eddy_parameters" % self.inputs.out_base
        )

        # File generation might depend on the version of EDDY
        out_rotated_bvecs = os.path.abspath(
            "%s.eddy_rotated_bvecs" % self.inputs.out_base
        )
        out_movement_rms = os.path.abspath(
            "%s.eddy_movement_rms" % self.inputs.out_base
        )
        out_restricted_movement_rms = os.path.abspath(
            "%s.eddy_restricted_movement_rms" % self.inputs.out_base
        )
        out_shell_alignment_parameters = os.path.abspath(
            "%s.eddy_post_eddy_shell_alignment_parameters" % self.inputs.out_base
        )
        out_shell_pe_translation_parameters = os.path.abspath(
            "%s.eddy_post_eddy_shell_PE_translation_parameters" % self.inputs.out_base
        )
        out_outlier_map = os.path.abspath("%s.eddy_outlier_map" % self.inputs.out_base)
        out_outlier_n_stdev_map = os.path.abspath(
            "%s.eddy_outlier_n_stdev_map" % self.inputs.out_base
        )
        out_outlier_n_sqr_stdev_map = os.path.abspath(
            "%s.eddy_outlier_n_sqr_stdev_map" % self.inputs.out_base
        )
        out_outlier_report = os.path.abspath(
            "%s.eddy_outlier_report" % self.inputs.out_base
        )
        if isdefined(self.inputs.repol) and self.inputs.repol:
            out_outlier_free = os.path.abspath(
                "%s.eddy_outlier_free_data" % self.inputs.out_base
            )
            if os.path.exists(out_outlier_free):
                outputs["out_outlier_free"] = out_outlier_free
        if isdefined(self.inputs.mporder) and self.inputs.mporder > 0:
            out_movement_over_time = os.path.abspath(
                "%s.eddy_movement_over_time" % self.inputs.out_base
            )
            if os.path.exists(out_movement_over_time):
                outputs["out_movement_over_time"] = out_movement_over_time
        if isdefined(self.inputs.cnr_maps) and self.inputs.cnr_maps:
            out_cnr_maps = os.path.abspath(
                "%s.eddy_cnr_maps.nii.gz" % self.inputs.out_base
            )
            if os.path.exists(out_cnr_maps):
                outputs["out_cnr_maps"] = out_cnr_maps
        if isdefined(self.inputs.residuals) and self.inputs.residuals:
            out_residuals = os.path.abspath(
                "%s.eddy_residuals.nii.gz" % self.inputs.out_base
            )
            if os.path.exists(out_residuals):
                outputs["out_residuals"] = out_residuals

        if os.path.exists(out_rotated_bvecs):
            outputs["out_rotated_bvecs"] = out_rotated_bvecs
        if os.path.exists(out_movement_rms):
            outputs["out_movement_rms"] = out_movement_rms
        if os.path.exists(out_restricted_movement_rms):
            outputs["out_restricted_movement_rms"] = out_restricted_movement_rms
        if os.path.exists(out_shell_alignment_parameters):
            outputs["out_shell_alignment_parameters"] = out_shell_alignment_parameters
        if os.path.exists(out_shell_pe_translation_parameters):
            outputs[
                "out_shell_pe_translation_parameters"
            ] = out_shell_pe_translation_parameters
        if os.path.exists(out_outlier_map):
            outputs["out_outlier_map"] = out_outlier_map
        if os.path.exists(out_outlier_n_stdev_map):
            outputs["out_outlier_n_stdev_map"] = out_outlier_n_stdev_map
        if os.path.exists(out_outlier_n_sqr_stdev_map):
            outputs["out_outlier_n_sqr_stdev_map"] = out_outlier_n_sqr_stdev_map
        if os.path.exists(out_outlier_report):
            outputs["out_outlier_report"] = out_outlier_report

        return outputs


class SigLossInputSpec(FSLCommandInputSpec):
    in_file = File(mandatory=True, exists=True, argstr="-i %s", desc="b0 fieldmap file")
    out_file = File(
        argstr="-s %s", desc="output signal loss estimate file", genfile=True
    )

    mask_file = File(exists=True, argstr="-m %s", desc="brain mask file")
    echo_time = traits.Float(argstr="--te=%f", desc="echo time in seconds")
    slice_direction = traits.Enum(
        "x", "y", "z", argstr="-d %s", desc="slicing direction"
    )


class SigLossOuputSpec(TraitedSpec):
    out_file = File(exists=True, desc="signal loss estimate file")


class SigLoss(FSLCommand):
    """
    Estimates signal loss from a field map (in rad/s)

    Examples
    --------

    >>> from nipype.interfaces.fsl import SigLoss
    >>> sigloss = SigLoss()
    >>> sigloss.inputs.in_file = "phase.nii"
    >>> sigloss.inputs.echo_time = 0.03
    >>> sigloss.inputs.output_type = "NIFTI_GZ"
    >>> sigloss.cmdline # doctest: +ELLIPSIS
    'sigloss --te=0.030000 -i phase.nii -s .../phase_sigloss.nii.gz'
    >>> res = sigloss.run() # doctest: +SKIP


    """

    input_spec = SigLossInputSpec
    output_spec = SigLossOuputSpec
    _cmd = "sigloss"

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_file
        if (not isdefined(outputs["out_file"])) and (isdefined(self.inputs.in_file)):
            outputs["out_file"] = self._gen_fname(
                self.inputs.in_file, suffix="_sigloss"
            )
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()["out_file"]
        return None


class EpiRegInputSpec(FSLCommandInputSpec):
    epi = File(
        exists=True, argstr="--epi=%s", mandatory=True, position=-4, desc="EPI image"
    )
    t1_head = File(
        exists=True,
        argstr="--t1=%s",
        mandatory=True,
        position=-3,
        desc="wholehead T1 image",
    )
    t1_brain = File(
        exists=True,
        argstr="--t1brain=%s",
        mandatory=True,
        position=-2,
        desc="brain extracted T1 image",
    )
    out_base = traits.String(
        "epi2struct",
        desc="output base name",
        argstr="--out=%s",
        position=-1,
        usedefault=True,
    )
    fmap = File(exists=True, argstr="--fmap=%s", desc="fieldmap image (in rad/s)")
    fmapmag = File(
        exists=True, argstr="--fmapmag=%s", desc="fieldmap magnitude image - wholehead"
    )
    fmapmagbrain = File(
        exists=True,
        argstr="--fmapmagbrain=%s",
        desc="fieldmap magnitude image - brain extracted",
    )
    wmseg = File(
        exists=True,
        argstr="--wmseg=%s",
        desc="white matter segmentation of T1 image, has to be named \
                 like the t1brain and end on _wmseg",
    )
    echospacing = traits.Float(
        argstr="--echospacing=%f",
        desc="Effective EPI echo spacing  \
                               (sometimes called dwell time) - in seconds",
    )
    pedir = traits.Enum(
        "x",
        "y",
        "z",
        "-x",
        "-y",
        "-z",
        argstr="--pedir=%s",
        desc="phase encoding direction, dir = x/y/z/-x/-y/-z",
    )

    weight_image = File(
        exists=True, argstr="--weight=%s", desc="weighting image (in T1 space)"
    )
    no_fmapreg = traits.Bool(
        False,
        argstr="--nofmapreg",
        desc="do not perform registration of fmap to T1 \
                        (use if fmap already registered)",
    )
    no_clean = traits.Bool(
        True,
        argstr="--noclean",
        usedefault=True,
        desc="do not clean up intermediate files",
    )


class EpiRegOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="unwarped and coregistered epi input")
    out_1vol = File(exists=True, desc="unwarped and coregistered single volume")
    fmap2str_mat = File(exists=True, desc="rigid fieldmap-to-structural transform")
    fmap2epi_mat = File(exists=True, desc="rigid fieldmap-to-epi transform")
    fmap_epi = File(exists=True, desc="fieldmap in epi space")
    fmap_str = File(exists=True, desc="fieldmap in structural space")
    fmapmag_str = File(exists=True, desc="fieldmap magnitude image in structural space")
    epi2str_inv = File(exists=True, desc="rigid structural-to-epi transform")
    epi2str_mat = File(exists=True, desc="rigid epi-to-structural transform")
    shiftmap = File(exists=True, desc="shiftmap in epi space")
    fullwarp = File(
        exists=True,
        desc="warpfield to unwarp epi and transform into \
                    structural space",
    )
    wmseg = File(exists=True, desc="white matter segmentation used in flirt bbr")
    seg = File(exists=True, desc="white matter, gray matter, csf segmentation")
    wmedge = File(exists=True, desc="white matter edges for visualization")


class EpiReg(FSLCommand):
    """

    Runs FSL epi_reg script for simultaneous coregistration and fieldmap
    unwarping.

    Examples
    --------

    >>> from nipype.interfaces.fsl import EpiReg
    >>> epireg = EpiReg()
    >>> epireg.inputs.epi='epi.nii'
    >>> epireg.inputs.t1_head='T1.nii'
    >>> epireg.inputs.t1_brain='T1_brain.nii'
    >>> epireg.inputs.out_base='epi2struct'
    >>> epireg.inputs.fmap='fieldmap_phase_fslprepared.nii'
    >>> epireg.inputs.fmapmag='fieldmap_mag.nii'
    >>> epireg.inputs.fmapmagbrain='fieldmap_mag_brain.nii'
    >>> epireg.inputs.echospacing=0.00067
    >>> epireg.inputs.pedir='y'
    >>> epireg.cmdline # doctest: +ELLIPSIS
    'epi_reg --echospacing=0.000670 --fmap=fieldmap_phase_fslprepared.nii \
--fmapmag=fieldmap_mag.nii --fmapmagbrain=fieldmap_mag_brain.nii --noclean \
--pedir=y --epi=epi.nii --t1=T1.nii --t1brain=T1_brain.nii --out=epi2struct'
    >>> epireg.run() # doctest: +SKIP

    """

    _cmd = "epi_reg"
    input_spec = EpiRegInputSpec
    output_spec = EpiRegOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.join(
            os.getcwd(), self.inputs.out_base + ".nii.gz"
        )
        if not (
            isdefined(self.inputs.no_fmapreg) and self.inputs.no_fmapreg
        ) and isdefined(self.inputs.fmap):
            outputs["out_1vol"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_1vol.nii.gz"
            )
            outputs["fmap2str_mat"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmap2str.mat"
            )
            outputs["fmap2epi_mat"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmaprads2epi.mat"
            )
            outputs["fmap_epi"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmaprads2epi.nii.gz"
            )
            outputs["fmap_str"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmaprads2str.nii.gz"
            )
            outputs["fmapmag_str"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmap2str.nii.gz"
            )
            outputs["shiftmap"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fieldmaprads2epi_shift.nii.gz"
            )
            outputs["fullwarp"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_warp.nii.gz"
            )
            outputs["epi2str_inv"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_inv.mat"
            )
        if not isdefined(self.inputs.wmseg):
            outputs["wmedge"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fast_wmedge.nii.gz"
            )
            outputs["wmseg"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fast_wmseg.nii.gz"
            )
            outputs["seg"] = os.path.join(
                os.getcwd(), self.inputs.out_base + "_fast_seg.nii.gz"
            )
        outputs["epi2str_mat"] = os.path.join(
            os.getcwd(), self.inputs.out_base + ".mat"
        )
        return outputs


#######################################
# deprecated interfaces
#######################################


class EPIDeWarpInputSpec(FSLCommandInputSpec):
    mag_file = File(
        exists=True,
        desc="Magnitude file",
        argstr="--mag %s",
        position=0,
        mandatory=True,
    )
    dph_file = File(
        exists=True,
        desc="Phase file assumed to be scaled from 0 to 4095",
        argstr="--dph %s",
        mandatory=True,
    )
    exf_file = File(
        exists=True, desc="example func volume (or use epi)", argstr="--exf %s"
    )
    epi_file = File(exists=True, desc="EPI volume to unwarp", argstr="--epi %s")
    tediff = traits.Float(
        2.46,
        usedefault=True,
        desc="difference in B0 field map TEs",
        argstr="--tediff %s",
    )
    esp = traits.Float(
        0.58, desc="EPI echo spacing", argstr="--esp %s", usedefault=True
    )
    sigma = traits.Int(
        2,
        usedefault=True,
        argstr="--sigma %s",
        desc="2D spatial gaussing smoothing \
                       stdev (default = 2mm)",
    )
    vsm = traits.String(genfile=True, desc="voxel shift map", argstr="--vsm %s")
    exfdw = traits.String(
        desc="dewarped example func volume", genfile=True, argstr="--exfdw %s"
    )
    epidw = traits.String(
        desc="dewarped epi volume", genfile=False, argstr="--epidw %s"
    )
    tmpdir = traits.String(genfile=True, desc="tmpdir", argstr="--tmpdir %s")
    nocleanup = traits.Bool(
        True, usedefault=True, desc="no cleanup", argstr="--nocleanup"
    )
    cleanup = traits.Bool(desc="cleanup", argstr="--cleanup")


class EPIDeWarpOutputSpec(TraitedSpec):
    unwarped_file = File(desc="unwarped epi file")
    vsm_file = File(desc="voxel shift map")
    exfdw = File(desc="dewarped functional volume example")
    exf_mask = File(desc="Mask from example functional volume")


class EPIDeWarp(FSLCommand):
    """
    Wraps the unwarping script `epidewarp.fsl
    <http://surfer.nmr.mgh.harvard.edu/fswiki/epidewarp.fsl>`_.

    .. warning:: deprecated in FSL, please use
      :func:`niflow.nipype1.workflows.dmri.preprocess.epi.sdc_fmb` instead.

    Examples
    --------

    >>> from nipype.interfaces.fsl import EPIDeWarp
    >>> dewarp = EPIDeWarp()
    >>> dewarp.inputs.epi_file = "functional.nii"
    >>> dewarp.inputs.mag_file = "magnitude.nii"
    >>> dewarp.inputs.dph_file = "phase.nii"
    >>> dewarp.inputs.output_type = "NIFTI_GZ"
    >>> dewarp.cmdline # doctest: +ELLIPSIS
    'epidewarp.fsl --mag magnitude.nii --dph phase.nii --epi functional.nii \
--esp 0.58 --exfdw .../exfdw.nii.gz --nocleanup --sigma 2 --tediff 2.46 \
--tmpdir .../temp --vsm .../vsm.nii.gz'
    >>> res = dewarp.run() # doctest: +SKIP


    """

    _cmd = "epidewarp.fsl"
    input_spec = EPIDeWarpInputSpec
    output_spec = EPIDeWarpOutputSpec

    def __init__(self, **inputs):
        warnings.warn(
            (
                "Deprecated: Please use "
                "niflow.nipype1.workflows.dmri.preprocess.epi.sdc_fmb instead"
            ),
            DeprecationWarning,
        )
        return super(EPIDeWarp, self).__init__(**inputs)

    def _run_interface(self, runtime):
        runtime = super(EPIDeWarp, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _gen_filename(self, name):
        if name == "exfdw":
            if isdefined(self.inputs.exf_file):
                return self._gen_fname(self.inputs.exf_file, suffix="_exfdw")
            else:
                return self._gen_fname("exfdw")
        if name == "epidw":
            if isdefined(self.inputs.epi_file):
                return self._gen_fname(self.inputs.epi_file, suffix="_epidw")
        if name == "vsm":
            return self._gen_fname("vsm")
        if name == "tmpdir":
            return os.path.join(os.getcwd(), "temp")
        return None

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if not isdefined(self.inputs.exfdw):
            outputs["exfdw"] = self._gen_filename("exfdw")
        else:
            outputs["exfdw"] = self.inputs.exfdw
        if isdefined(self.inputs.epi_file):
            if isdefined(self.inputs.epidw):
                outputs["unwarped_file"] = self.inputs.epidw
            else:
                outputs["unwarped_file"] = self._gen_filename("epidw")
        if not isdefined(self.inputs.vsm):
            outputs["vsm_file"] = self._gen_filename("vsm")
        else:
            outputs["vsm_file"] = self._gen_fname(self.inputs.vsm)
        if not isdefined(self.inputs.tmpdir):
            outputs["exf_mask"] = self._gen_fname(
                cwd=self._gen_filename("tmpdir"), basename="maskexf"
            )
        else:
            outputs["exf_mask"] = self._gen_fname(
                cwd=self.inputs.tmpdir, basename="maskexf"
            )
        return outputs


class EddyCorrectInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True, desc="4D input file", argstr="%s", position=0, mandatory=True
    )
    out_file = File(
        desc="4D output file",
        argstr="%s",
        position=1,
        name_source=["in_file"],
        name_template="%s_edc",
        output_name="eddy_corrected",
    )
    ref_num = traits.Int(
        0,
        argstr="%d",
        position=2,
        desc="reference number",
        mandatory=True,
        usedefault=True,
    )


class EddyCorrectOutputSpec(TraitedSpec):
    eddy_corrected = File(
        exists=True, desc="path/name of 4D eddy corrected output file"
    )


class EddyCorrect(FSLCommand):
    """

    .. warning:: Deprecated in FSL. Please use
      :class:`nipype.interfaces.fsl.epi.Eddy` instead

    Example
    -------

    >>> from nipype.interfaces.fsl import EddyCorrect
    >>> eddyc = EddyCorrect(in_file='diffusion.nii',
    ...                     out_file="diffusion_edc.nii", ref_num=0)
    >>> eddyc.cmdline
    'eddy_correct diffusion.nii diffusion_edc.nii 0'

    """

    _cmd = "eddy_correct"
    input_spec = EddyCorrectInputSpec
    output_spec = EddyCorrectOutputSpec

    def __init__(self, **inputs):
        warnings.warn(
            ("Deprecated: Please use nipype.interfaces.fsl.epi.Eddy " "instead"),
            DeprecationWarning,
        )
        return super(EddyCorrect, self).__init__(**inputs)

    def _run_interface(self, runtime):
        runtime = super(EddyCorrect, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime


class EddyQuadInputSpec(FSLCommandInputSpec):
    base_name = traits.Str(
        "eddy_corrected",
        usedefault=True,
        argstr="%s",
        desc=(
            "Basename (including path) for EDDY output files, i.e., "
            "corrected images and QC files"
        ),
        position=0,
    )
    idx_file = File(
        exists=True,
        mandatory=True,
        argstr="--eddyIdx %s",
        desc=("File containing indices for all volumes into acquisition " "parameters"),
    )
    param_file = File(
        exists=True,
        mandatory=True,
        argstr="--eddyParams %s",
        desc="File containing acquisition parameters",
    )
    mask_file = File(
        exists=True, mandatory=True, argstr="--mask %s", desc="Binary mask file"
    )
    bval_file = File(
        exists=True, mandatory=True, argstr="--bvals %s", desc="b-values file"
    )
    bvec_file = File(
        exists=True,
        argstr="--bvecs %s",
        desc=(
            "b-vectors file - only used when <base_name>.eddy_residuals "
            "file is present"
        ),
    )
    output_dir = traits.Str(
        name_template="%s.qc",
        name_source=["base_name"],
        argstr="--output-dir %s",
        desc="Output directory - default = '<base_name>.qc'",
    )
    field = File(exists=True, argstr="--field %s", desc="TOPUP estimated field (in Hz)")
    slice_spec = File(
        exists=True,
        argstr="--slspec %s",
        desc="Text file specifying slice/group acquisition",
    )
    verbose = traits.Bool(argstr="--verbose", desc="Display debug messages")


class EddyQuadOutputSpec(TraitedSpec):
    qc_json = File(
        exists=True,
        desc=("Single subject database containing quality metrics and data " "info."),
    )
    qc_pdf = File(exists=True, desc="Single subject QC report.")
    avg_b_png = traits.List(
        File(exists=True),
        desc=(
            "Image showing mid-sagittal, -coronal and -axial slices of "
            "each averaged b-shell volume."
        ),
    )
    avg_b0_pe_png = traits.List(
        File(exists=True),
        desc=(
            "Image showing mid-sagittal, -coronal and -axial slices of "
            "each averaged pe-direction b0 volume. Generated when using "
            "the -f option."
        ),
    )
    cnr_png = traits.List(
        File(exists=True),
        desc=(
            "Image showing mid-sagittal, -coronal and -axial slices of "
            "each b-shell CNR volume. Generated when CNR maps are "
            "available."
        ),
    )
    vdm_png = File(
        exists=True,
        desc=(
            "Image showing mid-sagittal, -coronal and -axial slices of "
            "the voxel displacement map. Generated when using the -f "
            "option."
        ),
    )
    residuals = File(
        exists=True,
        desc=(
            "Text file containing the volume-wise mask-averaged squared "
            "residuals. Generated when residual maps are available."
        ),
    )
    clean_volumes = File(
        exists=True,
        desc=(
            "Text file containing a list of clean volumes, based on "
            "the eddy squared residuals. To generate a version of the "
            "pre-processed dataset without outlier volumes, use: "
            "`fslselectvols -i <eddy_corrected_data> -o "
            "eddy_corrected_data_clean --vols=vols_no_outliers.txt`"
        ),
    )


class EddyQuad(FSLCommand):
    """
    Interface for FSL eddy_quad, a tool for generating single subject reports
    and storing the quality assessment indices for each subject.
    `User guide <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddyqc/UsersGuide>`__

    Examples
    --------

    >>> from nipype.interfaces.fsl import EddyQuad
    >>> quad = EddyQuad()
    >>> quad.inputs.base_name  = 'eddy_corrected'
    >>> quad.inputs.idx_file   = 'epi_index.txt'
    >>> quad.inputs.param_file = 'epi_acqp.txt'
    >>> quad.inputs.mask_file  = 'epi_mask.nii'
    >>> quad.inputs.bval_file  = 'bvals.scheme'
    >>> quad.inputs.bvec_file  = 'bvecs.scheme'
    >>> quad.inputs.output_dir = 'eddy_corrected.qc'
    >>> quad.inputs.field      = 'fieldmap_phase_fslprepared.nii'
    >>> quad.inputs.verbose    = True
    >>> quad.cmdline
    'eddy_quad eddy_corrected --bvals bvals.scheme --bvecs bvecs.scheme \
--field fieldmap_phase_fslprepared.nii --eddyIdx epi_index.txt \
--mask epi_mask.nii --output-dir eddy_corrected.qc --eddyParams epi_acqp.txt \
--verbose'
    >>> res = quad.run() # doctest: +SKIP

    """

    _cmd = "eddy_quad"
    input_spec = EddyQuadInputSpec
    output_spec = EddyQuadOutputSpec

    def _list_outputs(self):
        from glob import glob

        outputs = self.output_spec().get()

        # If the output directory isn't defined, the interface seems to use
        # the default but not set its value in `self.inputs.output_dir`
        if not isdefined(self.inputs.output_dir):
            out_dir = os.path.abspath(os.path.basename(self.inputs.base_name) + ".qc")
        else:
            out_dir = os.path.abspath(self.inputs.output_dir)

        outputs["qc_json"] = os.path.join(out_dir, "qc.json")
        outputs["qc_pdf"] = os.path.join(out_dir, "qc.pdf")

        # Grab all b* files here. This will also grab the b0_pe* files
        # as well, but only if the field input was provided. So we'll remove
        # them later in the next conditional.
        outputs["avg_b_png"] = sorted(glob(os.path.join(out_dir, "avg_b*.png")))

        if isdefined(self.inputs.field):
            outputs["avg_b0_pe_png"] = sorted(
                glob(os.path.join(out_dir, "avg_b0_pe*.png"))
            )

            # The previous glob for `avg_b_png` also grabbed the
            # `avg_b0_pe_png` files so we have to remove them
            # from `avg_b_png`.
            for fname in outputs["avg_b0_pe_png"]:
                outputs["avg_b_png"].remove(fname)

            outputs["vdm_png"] = os.path.join(out_dir, "vdm.png")

        outputs["cnr_png"] = sorted(glob(os.path.join(out_dir, "cnr*.png")))

        residuals = os.path.join(out_dir, "eddy_msr.txt")
        if os.path.isfile(residuals):
            outputs["residuals"] = residuals

        clean_volumes = os.path.join(out_dir, "vols_no_outliers.txt")
        if os.path.isfile(clean_volumes):
            outputs["clean_volumes"] = clean_volumes

        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..segmentation import CorticalThickness


def test_CorticalThickness_inputs():
    input_map = dict(
        anatomical_image=dict(
            argstr="-a %s",
            extensions=None,
            mandatory=True,
        ),
        args=dict(
            argstr="%s",
        ),
        b_spline_smoothing=dict(
            argstr="-v",
        ),
        brain_probability_mask=dict(
            argstr="-m %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        brain_template=dict(
            argstr="-e %s",
            extensions=None,
            mandatory=True,
        ),
        cortical_label_image=dict(
            extensions=None,
        ),
        debug=dict(
            argstr="-z 1",
        ),
        dimension=dict(
            argstr="-d %d",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        extraction_registration_mask=dict(
            argstr="-f %s",
            extensions=None,
        ),
        image_suffix=dict(
            argstr="-s %s",
            usedefault=True,
        ),
        keep_temporary_files=dict(
            argstr="-k %d",
        ),
        label_propagation=dict(
            argstr="-l %s",
        ),
        max_iterations=dict(
            argstr="-i %d",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_prefix=dict(
            argstr="-o %s",
            usedefault=True,
        ),
        posterior_formulation=dict(
            argstr="-b %s",
        ),
        prior_segmentation_weight=dict(
            argstr="-w %f",
        ),
        quick_registration=dict(
            argstr="-q 1",
        ),
        segmentation_iterations=dict(
            argstr="-n %d",
        ),
        segmentation_priors=dict(
            argstr="-p %s",
            mandatory=True,
        ),
        t1_registration_template=dict(
            argstr="-t %s",
            extensions=None,
            mandatory=True,
        ),
        use_floatingpoint_precision=dict(
            argstr="-j %d",
        ),
        use_random_seeding=dict(
            argstr="-u %d",
        ),
    )
    inputs = CorticalThickness.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CorticalThickness_outputs():
    output_map = dict(
        BrainExtractionMask=dict(
            extensions=None,
        ),
        BrainSegmentation=dict(
            extensions=None,
        ),
        BrainSegmentationN4=dict(
            extensions=None,
        ),
        BrainSegmentationPosteriors=dict(),
        BrainVolumes=dict(
            extensions=None,
        ),
        CorticalThickness=dict(
            extensions=None,
        ),
        CorticalThicknessNormedToTemplate=dict(
            extensions=None,
        ),
        ExtractedBrainN4=dict(
            extensions=None,
        ),
        SubjectToTemplate0GenericAffine=dict(
            extensions=None,
        ),
        SubjectToTemplate1Warp=dict(
            extensions=None,
        ),
        SubjectToTemplateLogJacobian=dict(
            extensions=None,
        ),
        TemplateToSubject0Warp=dict(
            extensions=None,
        ),
        TemplateToSubject1GenericAffine=dict(
            extensions=None,
        ),
    )
    outputs = CorticalThickness.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# Modified 2017.04.21 by Chris Markiewicz
import pytest

from ..base import FSSurfaceCommand
from ... import freesurfer as fs
from ...io import FreeSurferSource


def test_FSSurfaceCommand_inputs():
    input_map = dict(
        args=dict(argstr="%s"),
        environ=dict(nohash=True, usedefault=True),
        subjects_dir=dict(),
    )
    inputs = FSSurfaceCommand.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


@pytest.mark.skipif(fs.no_freesurfer(), reason="freesurfer is not installed")
def test_associated_file(tmpdir):
    fssrc = FreeSurferSource(
        subjects_dir=fs.Info.subjectsdir(), subject_id="fsaverage", hemi="lh"
    )
    fssrc.base_dir = tmpdir.strpath
    fssrc.resource_monitor = False

    fsavginfo = fssrc.run().outputs.get()

    # Pairs of white/pial files in the same directories
    for white, pial in [
        ("lh.white", "lh.pial"),
        ("./lh.white", "./lh.pial"),
        (fsavginfo["white"], fsavginfo["pial"]),
    ]:

        # Unspecified paths, possibly with missing hemisphere information,
        # are equivalent to using the same directory and hemisphere
        for name in ("pial", "lh.pial", pial):
            assert FSSurfaceCommand._associated_file(white, name) == pial

        # With path information, no changes are made
        for name in ("./pial", "./lh.pial", fsavginfo["pial"]):
            assert FSSurfaceCommand._associated_file(white, name) == name



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..nx import AverageNetworks


def test_AverageNetworks_inputs():
    input_map = dict(
        group_id=dict(
            usedefault=True,
        ),
        in_files=dict(
            mandatory=True,
        ),
        out_gexf_groupavg=dict(
            extensions=None,
        ),
        out_gpickled_groupavg=dict(
            extensions=None,
        ),
        resolution_network_file=dict(
            extensions=None,
        ),
    )
    inputs = AverageNetworks.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_AverageNetworks_outputs():
    output_map = dict(
        gexf_groupavg=dict(
            extensions=None,
        ),
        gpickled_groupavg=dict(
            extensions=None,
        ),
        matlab_groupavgs=dict(),
    )
    outputs = AverageNetworks.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Seg


def test_Seg_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bias_classes=dict(
            argstr="-bias_classes %s",
        ),
        bias_fwhm=dict(
            argstr="-bias_fwhm %f",
        ),
        blur_meth=dict(
            argstr="-blur_meth %s",
        ),
        bmrf=dict(
            argstr="-bmrf %f",
        ),
        classes=dict(
            argstr="-classes %s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-anat %s",
            copyfile=True,
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        main_N=dict(
            argstr="-main_N %d",
        ),
        mask=dict(
            argstr="-mask %s",
            mandatory=True,
            position=-2,
        ),
        mixfloor=dict(
            argstr="-mixfloor %f",
        ),
        mixfrac=dict(
            argstr="-mixfrac %s",
        ),
        prefix=dict(
            argstr="-prefix %s",
        ),
    )
    inputs = Seg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Seg_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = Seg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""Quickshear is a simple geometric defacing algorithm."""

from .base import CommandLineInputSpec, CommandLine, traits, TraitedSpec, File
from ..external.due import BibTeX


class QuickshearInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        position=1,
        argstr="%s",
        mandatory=True,
        desc="neuroimage to deface",
    )
    mask_file = File(
        exists=True, position=2, argstr="%s", desc="brain mask", mandatory=True
    )
    out_file = File(
        name_template="%s_defaced",
        name_source="in_file",
        position=3,
        argstr="%s",
        desc="defaced output image",
        keep_extension=True,
    )
    buff = traits.Int(
        position=4,
        argstr="%d",
        desc="buffer size (in voxels) between shearing " "plane and the brain",
    )


class QuickshearOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="defaced output image")


class Quickshear(CommandLine):
    """
    Quickshear is a simple geometric defacing algorithm

    Given an anatomical image and a reasonable brainmask, Quickshear estimates
    a shearing plane with the brain mask on one side and the face on the other,
    zeroing out the face side.

    >>> from nipype.interfaces.quickshear import Quickshear
    >>> qs = Quickshear(in_file='T1.nii', mask_file='brain_mask.nii')
    >>> qs.cmdline
    'quickshear T1.nii brain_mask.nii T1_defaced.nii'

    In the absence of a precomputed mask, a simple pipeline can be generated
    with any tool that generates brain masks:

    >>> from nipype.pipeline import engine as pe
    >>> from nipype.interfaces import utility as niu
    >>> from nipype.interfaces.fsl import BET
    >>> deface_wf = pe.Workflow('deface_wf')
    >>> inputnode = pe.Node(niu.IdentityInterface(['in_file']),
    ...                     name='inputnode')
    >>> outputnode = pe.Node(niu.IdentityInterface(['out_file']),
    ...                      name='outputnode')
    >>> bet = pe.Node(BET(mask=True), name='bet')
    >>> quickshear = pe.Node(Quickshear(), name='quickshear')
    >>> deface_wf.connect([
    ...     (inputnode, bet, [('in_file', 'in_file')]),
    ...     (inputnode, quickshear, [('in_file', 'in_file')]),
    ...     (bet, quickshear, [('mask_file', 'mask_file')]),
    ...     (quickshear, outputnode, [('out_file', 'out_file')]),
    ...     ])
    >>> inputnode.inputs.in_file = 'T1.nii'
    >>> res = deface_wf.run()  # doctest: +SKIP
    """

    _cmd = "quickshear"
    input_spec = QuickshearInputSpec
    output_spec = QuickshearOutputSpec

    _references = [
        {
            "entry": BibTeX(
                "@inproceedings{Schimke2011,"
                "address = {San Francisco},"
                "author = {Schimke, Nakeisha and Hale, John},"
                "booktitle = {Proceedings of the 2nd USENIX Conference on "
                "Health Security and Privacy},"
                "title = {{Quickshear Defacing for Neuroimages}},"
                "year = {2011},"
                "month = sep}"
            ),
            "tags": ["implementation"],
        }
    ]



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import PlotTimeSeries


def test_PlotTimeSeries_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="%s",
            mandatory=True,
            position=1,
        ),
        labels=dict(
            argstr="%s",
        ),
        legend_file=dict(
            argstr="--legend=%s",
            extensions=None,
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
        output_type=dict(),
        plot_finish=dict(
            argstr="--finish=%d",
            xor=("plot_range",),
        ),
        plot_range=dict(
            argstr="%s",
            xor=("plot_start", "plot_finish"),
        ),
        plot_size=dict(
            argstr="%s",
        ),
        plot_start=dict(
            argstr="--start=%d",
            xor=("plot_range",),
        ),
        sci_notation=dict(
            argstr="--sci",
        ),
        title=dict(
            argstr="%s",
        ),
        x_precision=dict(
            argstr="--precision=%d",
        ),
        x_units=dict(
            argstr="-u %d",
            usedefault=True,
        ),
        y_max=dict(
            argstr="--ymax=%.2f",
            xor=("y_range",),
        ),
        y_min=dict(
            argstr="--ymin=%.2f",
            xor=("y_range",),
        ),
        y_range=dict(
            argstr="%s",
            xor=("y_min", "y_max"),
        ),
    )
    inputs = PlotTimeSeries.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PlotTimeSeries_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = PlotTimeSeries.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
    # changing to temporary directories
    >>> tmp = getfixture('tmpdir')
    >>> old = tmp.chdir()
"""
import os
import re
import numpy as np

from ..base import (
    traits,
    TraitedSpec,
    DynamicTraitedSpec,
    File,
    Undefined,
    isdefined,
    OutputMultiPath,
    InputMultiPath,
    BaseInterface,
    BaseInterfaceInputSpec,
    Str,
    SimpleInterface,
)
from ..io import IOBase, add_traits
from ...utils.filemanip import ensure_list, copyfile, split_filename


class IdentityInterface(IOBase):
    """Basic interface class generates identity mappings

    Examples
    --------

    >>> from nipype.interfaces.utility import IdentityInterface
    >>> ii = IdentityInterface(fields=['a', 'b'], mandatory_inputs=False)
    >>> ii.inputs.a
    <undefined>

    >>> ii.inputs.a = 'foo'
    >>> out = ii._outputs()
    >>> out.a
    <undefined>

    >>> out = ii.run()
    >>> out.outputs.a
    'foo'

    >>> ii2 = IdentityInterface(fields=['a', 'b'], mandatory_inputs=True)
    >>> ii2.inputs.a = 'foo'
    >>> out = ii2.run() # doctest: +SKIP
    ValueError: IdentityInterface requires a value for input 'b' because it was listed in 'fields' Interface IdentityInterface failed to run.
    """

    input_spec = DynamicTraitedSpec
    output_spec = DynamicTraitedSpec

    def __init__(self, fields=None, mandatory_inputs=True, **inputs):
        super(IdentityInterface, self).__init__(**inputs)
        if fields is None or not fields:
            raise ValueError("Identity Interface fields must be a non-empty list")
        # Each input must be in the fields.
        for in_field in inputs:
            if in_field not in fields:
                raise ValueError(
                    "Identity Interface input is not in the fields: %s" % in_field
                )
        self._fields = fields
        self._mandatory_inputs = mandatory_inputs
        add_traits(self.inputs, fields)
        # Adding any traits wipes out all input values set in superclass initialization,
        # even it the trait is not in the add_traits argument. The work-around is to reset
        # the values after adding the traits.
        self.inputs.trait_set(**inputs)

    def _add_output_traits(self, base):
        return add_traits(base, self._fields)

    def _list_outputs(self):
        # manual mandatory inputs check
        if self._fields and self._mandatory_inputs:
            for key in self._fields:
                value = getattr(self.inputs, key)
                if not isdefined(value):
                    msg = (
                        "%s requires a value for input '%s' because it was listed in 'fields'. \
                    You can turn off mandatory inputs checking by passing mandatory_inputs = False to the constructor."
                        % (self.__class__.__name__, key)
                    )
                    raise ValueError(msg)

        outputs = self._outputs().get()
        for key in self._fields:
            val = getattr(self.inputs, key)
            if isdefined(val):
                outputs[key] = val
        return outputs


class MergeInputSpec(DynamicTraitedSpec, BaseInterfaceInputSpec):
    axis = traits.Enum(
        "vstack",
        "hstack",
        usedefault=True,
        desc="direction in which to merge, hstack requires same number of elements in each input",
    )
    no_flatten = traits.Bool(
        False,
        usedefault=True,
        desc="append to outlist instead of extending in vstack mode",
    )
    ravel_inputs = traits.Bool(
        False, usedefault=True, desc="ravel inputs when no_flatten is False"
    )


class MergeOutputSpec(TraitedSpec):
    out = traits.List(desc="Merged output")


def _ravel(in_val):
    if not isinstance(in_val, list):
        return in_val
    flat_list = []
    for val in in_val:
        raveled_val = _ravel(val)
        if isinstance(raveled_val, list):
            flat_list.extend(raveled_val)
        else:
            flat_list.append(raveled_val)
    return flat_list


class Merge(IOBase):
    """Basic interface class to merge inputs into a single list

    ``Merge(1)`` will merge a list of lists

    Examples
    --------

    >>> from nipype.interfaces.utility import Merge
    >>> mi = Merge(3)
    >>> mi.inputs.in1 = 1
    >>> mi.inputs.in2 = [2, 5]
    >>> mi.inputs.in3 = 3
    >>> out = mi.run()
    >>> out.outputs.out
    [1, 2, 5, 3]

    >>> merge = Merge(1)
    >>> merge.inputs.in1 = [1, [2, 5], 3]
    >>> out = merge.run()
    >>> out.outputs.out
    [1, [2, 5], 3]

    >>> merge = Merge(1)
    >>> merge.inputs.in1 = [1, [2, 5], 3]
    >>> merge.inputs.ravel_inputs = True
    >>> out = merge.run()
    >>> out.outputs.out
    [1, 2, 5, 3]

    >>> merge = Merge(1)
    >>> merge.inputs.in1 = [1, [2, 5], 3]
    >>> merge.inputs.no_flatten = True
    >>> out = merge.run()
    >>> out.outputs.out
    [[1, [2, 5], 3]]
    """

    input_spec = MergeInputSpec
    output_spec = MergeOutputSpec

    def __init__(self, numinputs=0, **inputs):
        super(Merge, self).__init__(**inputs)
        self._numinputs = numinputs
        if numinputs >= 1:
            input_names = ["in%d" % (i + 1) for i in range(numinputs)]
        else:
            input_names = []
        add_traits(self.inputs, input_names)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out = []

        if self._numinputs < 1:
            return outputs
        else:
            getval = lambda idx: getattr(self.inputs, "in%d" % (idx + 1))
            values = [
                getval(idx) for idx in range(self._numinputs) if isdefined(getval(idx))
            ]

        if self.inputs.axis == "vstack":
            for value in values:
                if isinstance(value, list) and not self.inputs.no_flatten:
                    out.extend(_ravel(value) if self.inputs.ravel_inputs else value)
                else:
                    out.append(value)
        else:
            lists = [ensure_list(val) for val in values]
            out = [[val[i] for val in lists] for i in range(len(lists[0]))]
        outputs["out"] = out
        return outputs


class RenameInputSpec(DynamicTraitedSpec):
    in_file = File(exists=True, mandatory=True, desc="file to rename")
    keep_ext = traits.Bool(
        desc=("Keep in_file extension, replace " "non-extension component of name")
    )
    format_string = Str(
        mandatory=True, desc="Python formatting string for output template"
    )
    parse_string = Str(
        desc="Python regexp parse string to define " "replacement inputs"
    )
    use_fullpath = traits.Bool(
        False, usedefault=True, desc="Use full path as input to regex parser"
    )


class RenameOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="softlink to original file with new name")


class Rename(SimpleInterface, IOBase):
    """Change the name of a file based on a mapped format string.

    To use additional inputs that will be defined at run-time, the class
    constructor must be called with the format template, and the fields
    identified will become inputs to the interface.

    Additionally, you may set the parse_string input, which will be run
    over the input filename with a regular expressions search, and will
    fill in additional input fields from matched groups. Fields set with
    inputs have precedence over fields filled in with the regexp match.

    Examples
    --------

    >>> from nipype.interfaces.utility import Rename
    >>> rename1 = Rename()
    >>> rename1.inputs.in_file = os.path.join(datadir, "zstat1.nii.gz") # datadir is a directory with exemplary files, defined in conftest.py
    >>> rename1.inputs.format_string = "Faces-Scenes.nii.gz"
    >>> res = rename1.run()          # doctest: +SKIP
    >>> res.outputs.out_file         # doctest: +SKIP
    'Faces-Scenes.nii.gz"            # doctest: +SKIP

    >>> rename2 = Rename(format_string="%(subject_id)s_func_run%(run)02d")
    >>> rename2.inputs.in_file = os.path.join(datadir, "functional.nii")
    >>> rename2.inputs.keep_ext = True
    >>> rename2.inputs.subject_id = "subj_201"
    >>> rename2.inputs.run = 2
    >>> res = rename2.run()          # doctest: +SKIP
    >>> res.outputs.out_file         # doctest: +SKIP
    'subj_201_func_run02.nii'        # doctest: +SKIP

    >>> rename3 = Rename(format_string="%(subject_id)s_%(seq)s_run%(run)02d.nii")
    >>> rename3.inputs.in_file = os.path.join(datadir, "func_epi_1_1.nii")
    >>> rename3.inputs.parse_string = r"func_(?P<seq>\w*)_.*"
    >>> rename3.inputs.subject_id = "subj_201"
    >>> rename3.inputs.run = 2
    >>> res = rename3.run()          # doctest: +SKIP
    >>> res.outputs.out_file         # doctest: +SKIP
    'subj_201_epi_run02.nii'         # doctest: +SKIP

    """

    input_spec = RenameInputSpec
    output_spec = RenameOutputSpec

    def __init__(self, format_string=None, **inputs):
        super(Rename, self).__init__(**inputs)
        if format_string is not None:
            self.inputs.format_string = format_string
            self.fmt_fields = re.findall(r"%\((.+?)\)", format_string)
            add_traits(self.inputs, self.fmt_fields)
        else:
            self.fmt_fields = []

    def _rename(self):
        fmt_dict = dict()
        if isdefined(self.inputs.parse_string):
            if isdefined(self.inputs.use_fullpath) and self.inputs.use_fullpath:
                m = re.search(self.inputs.parse_string, self.inputs.in_file)
            else:
                m = re.search(
                    self.inputs.parse_string, os.path.split(self.inputs.in_file)[1]
                )
            if m:
                fmt_dict.update(m.groupdict())
        for field in self.fmt_fields:
            val = getattr(self.inputs, field)
            if isdefined(val):
                fmt_dict[field] = getattr(self.inputs, field)
        if self.inputs.keep_ext:
            fmt_string = "".join(
                [self.inputs.format_string, split_filename(self.inputs.in_file)[2]]
            )
        else:
            fmt_string = self.inputs.format_string
        return fmt_string % fmt_dict

    def _run_interface(self, runtime):
        runtime.returncode = 0
        out_file = os.path.join(runtime.cwd, self._rename())
        _ = copyfile(self.inputs.in_file, out_file)
        self._results["out_file"] = out_file
        return runtime


class SplitInputSpec(BaseInterfaceInputSpec):
    inlist = traits.List(traits.Any, mandatory=True, desc="list of values to split")
    splits = traits.List(
        traits.Int,
        mandatory=True,
        desc="Number of outputs in each split - should add to number of inputs",
    )
    squeeze = traits.Bool(
        False, usedefault=True, desc="unfold one-element splits removing the list"
    )


class Split(IOBase):
    """Basic interface class to split lists into multiple outputs

    Examples
    --------

    >>> from nipype.interfaces.utility import Split
    >>> sp = Split()
    >>> _ = sp.inputs.trait_set(inlist=[1, 2, 3], splits=[2, 1])
    >>> out = sp.run()
    >>> out.outputs.out1
    [1, 2]

    """

    input_spec = SplitInputSpec
    output_spec = DynamicTraitedSpec

    def _add_output_traits(self, base):
        undefined_traits = {}
        for i in range(len(self.inputs.splits)):
            key = "out%d" % (i + 1)
            base.add_trait(key, traits.Any)
            undefined_traits[key] = Undefined
        base.trait_set(trait_change_notify=False, **undefined_traits)
        return base

    def _list_outputs(self):
        outputs = self._outputs().get()
        if isdefined(self.inputs.splits):
            if sum(self.inputs.splits) != len(self.inputs.inlist):
                raise RuntimeError("sum of splits != num of list elements")
            splits = [0]
            splits.extend(self.inputs.splits)
            splits = np.cumsum(splits)
            for i in range(len(splits) - 1):
                val = np.array(self.inputs.inlist, dtype=object)[
                    splits[i] : splits[i + 1]
                ].tolist()
                if self.inputs.squeeze and len(val) == 1:
                    val = val[0]
                outputs["out%d" % (i + 1)] = val
        return outputs


class SelectInputSpec(BaseInterfaceInputSpec):
    inlist = InputMultiPath(
        traits.Any, mandatory=True, desc="list of values to choose from"
    )
    index = InputMultiPath(
        traits.Int, mandatory=True, desc="0-based indices of values to choose"
    )


class SelectOutputSpec(TraitedSpec):
    out = OutputMultiPath(traits.Any, desc="list of selected values")


class Select(IOBase):
    """Basic interface class to select specific elements from a list

    Examples
    --------

    >>> from nipype.interfaces.utility import Select
    >>> sl = Select()
    >>> _ = sl.inputs.trait_set(inlist=[1, 2, 3, 4, 5], index=[3])
    >>> out = sl.run()
    >>> out.outputs.out
    4

    >>> _ = sl.inputs.trait_set(inlist=[1, 2, 3, 4, 5], index=[3, 4])
    >>> out = sl.run()
    >>> out.outputs.out
    [4, 5]

    """

    input_spec = SelectInputSpec
    output_spec = SelectOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        out = np.array(self.inputs.inlist, dtype=object)[
            np.array(self.inputs.index)
        ].tolist()
        outputs["out"] = out
        return outputs


class AssertEqualInputSpec(BaseInterfaceInputSpec):
    volume1 = File(exists=True, mandatory=True)
    volume2 = File(exists=True, mandatory=True)


class AssertEqual(BaseInterface):
    input_spec = AssertEqualInputSpec

    def _run_interface(self, runtime):
        import nibabel as nb

        data1 = np.asanyarray(nb.load(self.inputs.volume1))
        data2 = np.asanyarray(nb.load(self.inputs.volume2))

        if not np.array_equal(data1, data2):
            raise RuntimeError("Input images are not exactly equal")
        return runtime



# -*- coding: utf-8 -*-
from .segmentation import SimilarityIndex, BRAINSTalairach, BRAINSTalairachMask
from .utilities import (
    HistogramMatchingFilter,
    GenerateEdgeMapImage,
    GeneratePurePlugMask,
)
from .classify import BRAINSPosteriorToContinuousClass



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..modelgen import SpecifyModel


def test_SpecifyModel_inputs():
    input_map = dict(
        bids_amplitude_column=dict(),
        bids_condition_column=dict(
            usedefault=True,
        ),
        bids_event_file=dict(
            mandatory=True,
            xor=["subject_info", "event_files", "bids_event_file"],
        ),
        event_files=dict(
            mandatory=True,
            xor=["subject_info", "event_files", "bids_event_file"],
        ),
        functional_runs=dict(
            copyfile=False,
            mandatory=True,
        ),
        high_pass_filter_cutoff=dict(
            mandatory=True,
        ),
        input_units=dict(
            mandatory=True,
        ),
        outlier_files=dict(
            copyfile=False,
        ),
        parameter_source=dict(
            usedefault=True,
        ),
        realignment_parameters=dict(
            copyfile=False,
        ),
        subject_info=dict(
            mandatory=True,
            xor=["subject_info", "event_files", "bids_event_file"],
        ),
        time_repetition=dict(
            mandatory=True,
        ),
    )
    inputs = SpecifyModel.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SpecifyModel_outputs():
    output_map = dict(
        session_info=dict(),
    )
    outputs = SpecifyModel.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..diffusion import TractographyLabelMapSeeding


def test_TractographyLabelMapSeeding_inputs():
    input_map = dict(
        InputVolume=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        OutputFibers=dict(
            argstr="%s",
            hash_files=False,
            position=-1,
        ),
        args=dict(
            argstr="%s",
        ),
        clthreshold=dict(
            argstr="--clthreshold %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputroi=dict(
            argstr="--inputroi %s",
            extensions=None,
        ),
        integrationsteplength=dict(
            argstr="--integrationsteplength %f",
        ),
        label=dict(
            argstr="--label %d",
        ),
        maximumlength=dict(
            argstr="--maximumlength %f",
        ),
        minimumlength=dict(
            argstr="--minimumlength %f",
        ),
        name=dict(
            argstr="--name %s",
        ),
        outputdirectory=dict(
            argstr="--outputdirectory %s",
            hash_files=False,
        ),
        randomgrid=dict(
            argstr="--randomgrid ",
        ),
        seedspacing=dict(
            argstr="--seedspacing %f",
        ),
        stoppingcurvature=dict(
            argstr="--stoppingcurvature %f",
        ),
        stoppingmode=dict(
            argstr="--stoppingmode %s",
        ),
        stoppingvalue=dict(
            argstr="--stoppingvalue %f",
        ),
        useindexspace=dict(
            argstr="--useindexspace ",
        ),
        writetofile=dict(
            argstr="--writetofile ",
        ),
    )
    inputs = TractographyLabelMapSeeding.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TractographyLabelMapSeeding_outputs():
    output_map = dict(
        OutputFibers=dict(
            extensions=None,
            position=-1,
        ),
        outputdirectory=dict(),
    )
    outputs = TractographyLabelMapSeeding.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class DWIUnbiasedNonLocalMeansFilterInputSpec(CommandLineInputSpec):
    rs = InputMultiPath(
        traits.Int,
        desc="The algorithm search for similar voxels in a neighborhood of this size (larger sizes than the default one are extremely slow).",
        sep=",",
        argstr="--rs %s",
    )
    rc = InputMultiPath(
        traits.Int,
        desc="Similarity between blocks is measured using windows of this size.",
        sep=",",
        argstr="--rc %s",
    )
    hp = traits.Float(
        desc="This parameter is related to noise; the larger the parameter, the more agressive the filtering. Should be near 1, and only values between 0.8 and 1.2 are allowed",
        argstr="--hp %f",
    )
    ng = traits.Int(
        desc="The number of the closest gradients that are used to jointly filter a given gradient direction (a maximum of 5 is allowed).",
        argstr="--ng %d",
    )
    re = InputMultiPath(
        traits.Int,
        desc="A neighborhood of this size is used to compute the statistics for noise estimation.",
        sep=",",
        argstr="--re %s",
    )
    inputVolume = File(position=-2, desc="Input DWI volume.", exists=True, argstr="%s")
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Output DWI volume.",
        argstr="%s",
    )


class DWIUnbiasedNonLocalMeansFilterOutputSpec(TraitedSpec):
    outputVolume = File(position=-1, desc="Output DWI volume.", exists=True)


class DWIUnbiasedNonLocalMeansFilter(SEMLikeCommandLine):
    """title: DWI Unbiased Non Local Means Filter

    category: Legacy.Diffusion.Denoising

    description: This module reduces noise (or unwanted detail) on a set of diffusion weighted images. For this, it filters the images using a Unbiased Non Local Means for Rician noise algorithm. It exploits not only the spatial redundancy, but the redundancy in similar gradient directions as well; it takes into account the N closest gradient directions to the direction being processed (a maximum of 5 gradient directions is allowed to keep a reasonable computational load, since we do not use neither similarity maps nor block-wise implementation).
    The noise parameter is automatically estimated in the same way as in the jointLMMSE module.
    A complete description of the algorithm may be found in:
    Antonio Tristan-Vega and Santiago Aja-Fernandez, DWI filtering using joint information for DTI and HARDI, Medical Image Analysis, Volume 14, Issue 2, Pages 205-218. 2010.
    Please, note that the execution of this filter is extremely slow, son only very conservative parameters (block size and search size as small as possible) should be used. Even so, its execution may take several hours. The advantage of this filter over joint LMMSE is its better preservation of edges and fine structures.

    version: 0.0.1.$Revision: 1 $(alpha)

    documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/UnbiasedNonLocalMeansFilterForDWI

    contributor: Antonio Tristan Vega (UVa), Santiago Aja Fernandez (UVa)

    acknowledgements: Partially founded by grant number TEC2007-67073/TCM from the Comision Interministerial de Ciencia y Tecnologia (Spain).
    """

    input_spec = DWIUnbiasedNonLocalMeansFilterInputSpec
    output_spec = DWIUnbiasedNonLocalMeansFilterOutputSpec
    _cmd = "DWIUnbiasedNonLocalMeansFilter "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..arithmetic import AddScalarVolumes


def test_AddScalarVolumes_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputVolume1=dict(
            argstr="%s",
            extensions=None,
            position=-3,
        ),
        inputVolume2=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        order=dict(
            argstr="--order %s",
        ),
        outputVolume=dict(
            argstr="%s",
            hash_files=False,
            position=-1,
        ),
    )
    inputs = AddScalarVolumes.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_AddScalarVolumes_outputs():
    output_map = dict(
        outputVolume=dict(
            extensions=None,
            position=-1,
        ),
    )
    outputs = AddScalarVolumes.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..fix import TrainingSetCreator


def test_TrainingSetCreator_inputs():
    input_map = dict(
        mel_icas_in=dict(
            argstr="%s",
            copyfile=False,
            position=-1,
        ),
    )
    inputs = TrainingSetCreator.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TrainingSetCreator_outputs():
    output_map = dict(
        mel_icas_out=dict(
            argstr="%s",
            copyfile=False,
            position=-1,
        ),
    )
    outputs = TrainingSetCreator.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dti import TractSkeleton


def test_TractSkeleton_inputs():
    input_map = dict(
        alt_data_file=dict(
            argstr="-a %s",
            extensions=None,
        ),
        alt_skeleton=dict(
            argstr="-s %s",
            extensions=None,
        ),
        args=dict(
            argstr="%s",
        ),
        data_file=dict(
            extensions=None,
        ),
        distance_map=dict(
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        output_type=dict(),
        project_data=dict(
            argstr="-p %.3f %s %s %s %s",
            requires=["threshold", "distance_map", "data_file"],
        ),
        projected_data=dict(
            extensions=None,
        ),
        search_mask_file=dict(
            extensions=None,
            xor=["use_cingulum_mask"],
        ),
        skeleton_file=dict(
            argstr="-o %s",
        ),
        threshold=dict(),
        use_cingulum_mask=dict(
            usedefault=True,
            xor=["search_mask_file"],
        ),
    )
    inputs = TractSkeleton.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TractSkeleton_outputs():
    output_map = dict(
        projected_data=dict(
            extensions=None,
        ),
        skeleton_file=dict(
            extensions=None,
        ),
    )
    outputs = TractSkeleton.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from .diffusion import dtiaverage, dtiestim, dtiprocess, DWIConvert
from .tractography import *
from .gtract import (
    gtractTransformToDisplacementField,
    gtractInvertBSplineTransform,
    gtractConcatDwi,
    gtractAverageBvalues,
    gtractCoregBvalues,
    gtractResampleAnisotropy,
    gtractResampleCodeImage,
    gtractCopyImageOrientation,
    gtractCreateGuideFiber,
    gtractAnisotropyMap,
    gtractClipAnisotropy,
    gtractResampleB0,
    gtractInvertRigidTransform,
    gtractImageConformity,
    compareTractInclusion,
    gtractFastMarchingTracking,
    gtractInvertDisplacementField,
    gtractCoRegAnatomy,
    gtractResampleDWIInPlace,
    gtractCostFastMarching,
    gtractFiberTracking,
    extractNrrdVectorIndex,
    gtractResampleFibers,
    gtractTensor,
)
from .maxcurvature import maxcurvature



# -*- coding: utf-8 -*-
from nipype.interfaces.fsl.model import FILMGLS, FILMGLSInputSpec


def test_filmgls():
    input_map = dict(
        args=dict(argstr="%s"),
        autocorr_estimate_only=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="-ac",
        ),
        autocorr_noestimate=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="-noest",
        ),
        brightness_threshold=dict(argstr="-epith %d"),
        design_file=dict(argstr="%s"),
        environ=dict(usedefault=True),
        fit_armodel=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="-ar",
        ),
        full_data=dict(argstr="-v"),
        in_file=dict(mandatory=True, argstr="%s"),
        mask_size=dict(argstr="-ms %d"),
        multitaper_product=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="-mt %d",
        ),
        output_pwdata=dict(argstr="-output_pwdata"),
        output_type=dict(),
        results_dir=dict(usedefault=True, argstr="-rn %s"),
        smooth_autocorr=dict(argstr="-sa"),
        threshold=dict(argstr="%f"),
        tukey_window=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="-tukey %d",
        ),
        use_pava=dict(argstr="-pava"),
    )
    input_map2 = dict(
        args=dict(argstr="%s"),
        autocorr_estimate_only=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="--ac",
        ),
        autocorr_noestimate=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="--noest",
        ),
        brightness_threshold=dict(argstr="--epith=%d"),
        design_file=dict(argstr="--pd=%s"),
        environ=dict(usedefault=True),
        fit_armodel=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="--ar",
        ),
        full_data=dict(argstr="-v"),
        in_file=dict(mandatory=True, argstr="--in=%s"),
        mask_size=dict(argstr="--ms=%d"),
        multitaper_product=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="--mt=%d",
        ),
        output_pwdata=dict(argstr="--outputPWdata"),
        output_type=dict(),
        results_dir=dict(argstr="--rn=%s", usedefault=True),
        smooth_autocorr=dict(argstr="--sa"),
        threshold=dict(usedefault=True, argstr="--thr=%f"),
        tukey_window=dict(
            xor=[
                "autocorr_estimate_only",
                "fit_armodel",
                "tukey_window",
                "multitaper_product",
                "use_pava",
                "autocorr_noestimate",
            ],
            argstr="--tukey=%d",
        ),
        use_pava=dict(argstr="--pava"),
    )
    instance = FILMGLS()
    if isinstance(instance.inputs, FILMGLSInputSpec):
        for key, metadata in list(input_map.items()):
            for metakey, value in list(metadata.items()):
                assert getattr(instance.inputs.traits()[key], metakey) == value
    else:
        for key, metadata in list(input_map2.items()):
            for metakey, value in list(metadata.items()):
                assert getattr(instance.inputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..developer import JistIntensityMp2rageMasking


def test_JistIntensityMp2rageMasking_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inBackground=dict(
            argstr="--inBackground %s",
        ),
        inMasking=dict(
            argstr="--inMasking %s",
        ),
        inQuantitative=dict(
            argstr="--inQuantitative %s",
            extensions=None,
        ),
        inSecond=dict(
            argstr="--inSecond %s",
            extensions=None,
        ),
        inSkip=dict(
            argstr="--inSkip %s",
        ),
        inT1weighted=dict(
            argstr="--inT1weighted %s",
            extensions=None,
        ),
        null=dict(
            argstr="--null %s",
        ),
        outMasked=dict(
            argstr="--outMasked_T1_Map %s",
            hash_files=False,
        ),
        outMasked2=dict(
            argstr="--outMasked_T1weighted %s",
            hash_files=False,
        ),
        outSignal=dict(
            argstr="--outSignal_Proba %s",
            hash_files=False,
        ),
        outSignal2=dict(
            argstr="--outSignal_Mask %s",
            hash_files=False,
        ),
        xDefaultMem=dict(
            argstr="-xDefaultMem %d",
        ),
        xMaxProcess=dict(
            argstr="-xMaxProcess %d",
            usedefault=True,
        ),
        xPrefExt=dict(
            argstr="--xPrefExt %s",
        ),
    )
    inputs = JistIntensityMp2rageMasking.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_JistIntensityMp2rageMasking_outputs():
    output_map = dict(
        outMasked=dict(
            extensions=None,
        ),
        outMasked2=dict(
            extensions=None,
        ),
        outSignal=dict(
            extensions=None,
        ),
        outSignal2=dict(
            extensions=None,
        ),
    )
    outputs = JistIntensityMp2rageMasking.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..convert import FSL2Scheme


def test_FSL2Scheme_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bscale=dict(
            argstr="-bscale %d",
            units="NA",
        ),
        bval_file=dict(
            argstr="-bvalfile %s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        bvec_file=dict(
            argstr="-bvecfile %s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        diffusiontime=dict(
            argstr="-diffusiontime %f",
            units="NA",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flipx=dict(
            argstr="-flipx",
        ),
        flipy=dict(
            argstr="-flipy",
        ),
        flipz=dict(
            argstr="-flipz",
        ),
        interleave=dict(
            argstr="-interleave",
        ),
        numscans=dict(
            argstr="-numscans %d",
            units="NA",
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        usegradmod=dict(
            argstr="-usegradmod",
        ),
    )
    inputs = FSL2Scheme.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FSL2Scheme_outputs():
    output_map = dict(
        scheme=dict(
            extensions=None,
        ),
    )
    outputs = FSL2Scheme.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import TCorr1D


def test_TCorr1D_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ktaub=dict(
            argstr=" -ktaub",
            position=1,
            xor=["pearson", "spearman", "quadrant"],
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            keep_extension=True,
            name_source="xset",
            name_template="%s_correlation.nii.gz",
        ),
        outputtype=dict(),
        pearson=dict(
            argstr=" -pearson",
            position=1,
            xor=["spearman", "quadrant", "ktaub"],
        ),
        quadrant=dict(
            argstr=" -quadrant",
            position=1,
            xor=["pearson", "spearman", "ktaub"],
        ),
        spearman=dict(
            argstr=" -spearman",
            position=1,
            xor=["pearson", "quadrant", "ktaub"],
        ),
        xset=dict(
            argstr=" %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        y_1d=dict(
            argstr=" %s",
            extensions=None,
            mandatory=True,
            position=-1,
        ),
    )
    inputs = TCorr1D.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TCorr1D_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = TCorr1D.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
